from datetime import datetime
from pathlib import Path

from pycrashreport.crash_report import Frame, Register, get_crash_report_from_file, BugType


def test_non_symbolicated_ios14():
    filename = Path(__file__).parent / 'user_mode_crash_report_ios14_non_symbolicated_abort.ips'
    crash_report = get_crash_report_from_file(open(filename, 'rt'))
    assert crash_report.bug_type == BugType.Crash_109
    assert crash_report.incident_id == '13917FF0-E1B1-4652-84C2-85516D101DFE'
    assert crash_report.timestamp == datetime(2021, 10, 22, 0, 14, 53)
    assert crash_report.faulting_thread == 7
    assert crash_report.exception_type == 'EXC_CRASH (SIGABRT)'
    assert crash_report.exception_subtype is None
    assert crash_report.application_specific_information == 'abort() called'

    expected_registers = [
        Register(name='x0', value=0),
        Register(name='x1', value=0),
        Register(name='x2', value=0),
        Register(name='x3', value=0),
        Register(name='x4', value=0x000000016f6763d0),
        Register(name='x5', value=0x000000016f676970),
        Register(name='x6', value=0x0000000000000072),
        Register(name='x7', value=0x0000000000001800),
        Register(name='x8', value=0xe3e68c37e41b1559),
        Register(name='x9', value=0xe3f5efb68b7c6559),
        Register(name='x10', value=0x0000000000000002),
        Register(name='x11', value=0x0000000000000003),
        Register(name='x12', value=0),
        Register(name='x13', value=0x00000000ffffffff),
        Register(name='x14', value=0x0000000000000010),
        Register(name='x15', value=0),
        Register(name='x16', value=0x0000000000000148),
        Register(name='x17', value=0x000000016f677000),
        Register(name='x18', value=0),
        Register(name='x19', value=0x0000000000000006),
        Register(name='x20', value=0x0000000000001f3b),
        Register(name='x21', value=0x000000016f6770e0),
        Register(name='x22', value=0x0000000000000114),
        Register(name='x23', value=0),
        Register(name='x24', value=0),
        Register(name='x25', value=0x000000016f6770e0),
        Register(name='x26', value=0),
        Register(name='x27', value=0x000000016f677180),
        Register(name='x28', value=0x00000000000003ff),
        Register(name='fp', value=0x000000016f6768e0),
        Register(name='lr', value=0x00000001e18a1a9c),
        Register(name='sp', value=0x000000016f6768c0),
        Register(name='pc', value=0x00000001c3e1a334),
        Register(name='cpsr', value=0x40000000),
        Register(name='esr', value=0x0000000056000080),
    ]
    for i, register in enumerate(expected_registers):
        assert crash_report.registers[i] == register

    expected_frames = [
        Frame(image_name='libsystem_kernel.dylib', image_base=0x1c3df1000, symbol=None,
              image_offset=168756, symbol_offset=None),
        Frame(image_name='libsystem_pthread.dylib', image_base=0x1e189f000, symbol=None,
              image_offset=10908, symbol_offset=None),
        Frame(image_name='libsystem_c.dylib', image_base=0x19ef2e000, symbol=None, image_offset=490372,
              symbol_offset=None),
        Frame(image_name='libc++abi.dylib', image_base=0x1aa804000, symbol=None, image_offset=80824,
              symbol_offset=None),
        Frame(image_name='libc++abi.dylib', image_base=0x1aa804000, symbol=None, image_offset=20168,
              symbol_offset=None),
        Frame(image_name='libobjc.A.dylib', image_base=0x1aa70e000, symbol=None, image_offset=28764,
              symbol_offset=None),
        Frame(image_name='libc++abi.dylib', image_base=0x1aa804000, symbol=None, image_offset=77728,
              symbol_offset=None),
        Frame(image_name='libc++abi.dylib', image_base=0x1aa804000, symbol=None, image_offset=77612,
              symbol_offset=None),
        Frame(image_name='libdispatch.dylib', image_base=0x1957c7000, symbol=None, image_offset=18480,
              symbol_offset=None),
        Frame(image_name='libdispatch.dylib', image_base=0x1957c7000, symbol=None, image_offset=31988,
              symbol_offset=None),
        Frame(image_name='libdispatch.dylib', image_base=0x1957c7000, symbol=None, image_offset=29572,
              symbol_offset=None),
        Frame(image_name='libdispatch.dylib', image_base=0x1957c7000, symbol=None, image_offset=90080,
              symbol_offset=None),
        Frame(image_name='libdispatch.dylib', image_base=0x1957c7000, symbol=None, image_offset=92120,
              symbol_offset=None),
        Frame(image_name='libsystem_pthread.dylib', image_base=0x1e189f000, symbol=None,
              image_offset=14184, symbol_offset=None),
        Frame(image_name='libsystem_pthread.dylib', image_base=0x1e189f000, symbol=None,
              image_offset=42828, symbol_offset=None),

    ]
    for i, frame in enumerate(expected_frames):
        assert frame == crash_report.frames[i]


def test_symbolicated_ios14():
    filename = str(Path(__file__).parent / 'user_mode_crash_report_ios14_symbolicated.ips')
    crash_report = get_crash_report_from_file(open(filename, 'rt'))
    assert crash_report.bug_type == BugType.Crash_109
    assert crash_report.incident_id == '2416C26A-72A8-4687-AFAA-7FCEB9D77458'
    assert crash_report.timestamp == datetime(2022, 1, 21, 18, 14, 26)
    assert crash_report.faulting_thread == 0
    assert crash_report.exception_type == 'EXC_CRASH (SIGABRT)'
    assert crash_report.exception_subtype is None
    assert crash_report.application_specific_information == 'dyld3 mode\nstack buffer overflow'

    expected_registers = [
        Register(name='x0', value=0x0000000000000000),
        Register(name='x1', value=0x0000000000000000),
        Register(name='x2', value=0x0000000000000000),
        Register(name='x3', value=0x0000000000000000),
        Register(name='x4', value=0x0000000000000000),
        Register(name='x5', value=0x0000000000000000),
        Register(name='x6', value=0x00676f6c7379732f),
        Register(name='x7', value=0xffffffffffffb5dc),
        Register(name='x8', value=0x000000010016f880),
        Register(name='x9', value=0xe021f14ad5b95ac2),
        Register(name='x10', value=0x0000000000000000),
        Register(name='x11', value=0x0000000000000038),
        Register(name='x12', value=0x00000001e5bf86c0),
        Register(name='x13', value=0x0000000089bff7fb),
        Register(name='x14', value=0x0000000000000001),
        Register(name='x15', value=0x00000000001ff800),
        Register(name='x16', value=0x0000000000000148),
        Register(name='x17', value=0x0000030100000380),
        Register(name='x18', value=0x0000000000000000),
        Register(name='x19', value=0x0000000000000006),
        Register(name='x20', value=0x0000000000000103),
        Register(name='x21', value=0x000000010016f960),
        Register(name='x22', value=0x0000000000000000),
        Register(name='x23', value=0x0000000000000000),
        Register(name='x24', value=0x0000000000000000),
        Register(name='x25', value=0x0000000000000000),
        Register(name='x26', value=0x0000000000000000),
        Register(name='x27', value=0x0000000000000000),
        Register(name='x28', value=0x000000016fd83af8),
        Register(name='fp', value=0x000000016fd839f0),
        Register(name='lr', value=0x00000001e5c049c4),
        Register(name='sp', value=0x000000016fd839d0),
        Register(name='pc', value=0x00000001c95957b0),
        Register(name='cpsr', value=0x40000000),
        Register(name='esr', value=0x56000080),
    ]
    for i, register in enumerate(expected_registers):
        assert crash_report.registers[i] == register

    expected_frames = [
        Frame(image_name='libsystem_kernel.dylib', image_base=None, symbol='__pthread_kill',
              image_offset=None, symbol_offset=8),
        Frame(image_name='libsystem_pthread.dylib', image_base=None, symbol='pthread_kill',
              image_offset=None, symbol_offset=212),
        Frame(image_name='libsystem_c.dylib', image_base=None, symbol='__abort', image_offset=None,
              symbol_offset=112),
        Frame(image_name='libsystem_c.dylib', image_base=None, symbol='a64l', image_offset=None,
              symbol_offset=0),
        Frame(image_name='kaki', image_base=None, symbol='main', image_offset=None,
              symbol_offset=108),
        Frame(image_name='libdyld.dylib', image_base=None, symbol='start', image_offset=None,
              symbol_offset=4),

    ]
    for i, frame in enumerate(expected_frames):
        assert frame == crash_report.frames[i]


def test_non_symbolicated_monterey():
    filename = str(Path(__file__).parent / 'user_mode_crash_report_monterey_non_symbolicated.ips')
    crash_report = get_crash_report_from_file(open(filename, 'rt'))
    assert crash_report.bug_type == BugType.Crash_309
    assert crash_report.incident_id == '051760D9-97FF-475F-8B61-B0FDFB04D484'
    assert crash_report.timestamp == datetime(2022, 1, 6, 15, 9, 22)
    assert crash_report.faulting_thread == 0
    assert crash_report.exception_type == 'EXC_BAD_ACCESS'
    assert crash_report.exception_subtype == 'KERN_INVALID_ADDRESS at 0x0000000000000000'
    assert crash_report.application_specific_information is None

    expected_registers = [
        Register(name='r13', value=140701961193192),
        Register(name='rax', value=4),
        Register(name='rflags', value=66119),
        Register(name='cpu', value=8),
        Register(name='r14', value=140701961192952),
        Register(name='rsi', value=0),
        Register(name='r8', value=32),
        Register(name='cr2', value=0),
        Register(name='rdx', value=1),
        Register(name='r10', value=1),
        Register(name='r9', value=187692599),
        Register(name='r15', value=140701961192952),
        Register(name='rbx', value=0),
        Register(name='trap', value=14),
        Register(name='err', value=20),
        Register(name='r11', value=582),
        Register(name='rip', value=0),
        Register(name='rbp', value=140701961192928),
        Register(name='rsp', value=140701961192872),
        Register(name='r12', value=4630279072),
        Register(name='rcx', value=140701961192872),
        Register(name='rdi', value=5123),
    ]
    for i, register in enumerate(expected_registers):
        assert crash_report.registers[i] == register

    expected_frames = [
        Frame(image_name=None, image_base=0, symbol=None, image_offset=0, symbol_offset=None),
        Frame(image_name='/usr/lib/system/libsystem_c.dylib', image_base=0x7ff80c65c000, symbol='nanosleep',
              image_offset=0x108a9, symbol_offset=0xc4),
        Frame(image_name='/bin/sleep', image_base=0x105857000, symbol=None, image_offset=0x3dd2, symbol_offset=None),
        Frame(image_name='/usr/lib/dyld', image_base=0x113f47000, symbol='start', image_offset=0x54fe,
              symbol_offset=0x1ce),
    ]
    for i, frame in enumerate(expected_frames):
        assert frame == crash_report.frames[i]


def test_forceReset_ios16():
    filename = str(Path(__file__).parent / 'kernel_mode_crash_report_ios16_forceReset-full.ips')
    crash_report = get_crash_report_from_file(open(filename, 'rt'))
    assert crash_report.bug_type == BugType.ForceReset
    assert crash_report.incident_id == '35F77863-C28D-42BA-B633-9732EA1F342A'
    assert crash_report.timestamp == datetime(2022, 12, 24, 11, 43, 0, 470000)



import click

from pycrashreport.crash_report import get_crash_report_from_file


@click.command()
@click.argument('file', type=click.File('rt'))
def cli(file):
    print(get_crash_report_from_file(file))


if __name__ == '__main__':
    cli()



import json
import posixpath
from collections import namedtuple
from datetime import datetime
from enum import Enum
from io import StringIO
from typing import List, Optional, Mapping, IO

import click
from cached_property import cached_property
from la_panic.panic_parser.kernel_panic import KernelPanic

Frame = namedtuple('Frame', 'image_name image_base image_offset symbol symbol_offset')
Register = namedtuple('Register', 'name value')


class BugType(Enum):
    WatchdogTimeout = '28'
    BasebandStats = '195'
    GPUEvent = '284'
    Sandbox = '187'
    TerminatingStackshot = '509'
    ServiceWatchdogTimeout = '29'
    Session = '179'
    LegacyStackshot = '188'
    MACorrelation = '197'
    iMessages = '189'
    log_power = '278'
    PowerLog = 'powerlog'
    DuetKnowledgeCollector2 = '58'
    BridgeRestore = '83'
    LegacyJetsam = '198'
    ExcResource_385 = '385'
    Modem = '199'
    Stackshot = '288'
    SystemInformation = 'system_profile'
    Jetsam_298 = '298'
    MemoryResource = '30'
    Bridge = '31'
    DifferentialPrivacy = 'diff_privacy'
    FirmwareIntegrity = '32'
    CoreAnalytics_33 = '33'
    AutoBugCapture = '34'
    EfiFirmwareIntegrity = '35'
    SystemStats = '36'
    AnonSystemStats = '37'
    Crash_9 = '9'
    Jetsam_98 = '98'
    LDCM = '100'
    Panic_10 = '10'
    Spin = '11'
    CLTM = '101'
    Hang = '12'
    Panic_110 = '110'
    ConnectionFailure = '13'
    MessageTracer = '14'
    LowBattery = '120'
    Siri = '201'
    ShutdownStall = '17'
    Panic_210 = '210'
    SymptomsCPUUsage = '202'
    AssumptionViolation = '18'
    CoreHandwriting = 'chw'
    IOMicroStackShot = '44'
    CoreAnalytics_211 = '211'
    SiriAppPrediction = '203'
    spin_45 = '45'
    PowerMicroStackshots = '220'
    BTMetadata = '212'
    SystemMemoryReset = '301'
    ResetCount = '115'
    AutoBugCapture_204 = '204'
    WifiCrashBinary = '221'
    MicroRunloopHang = '310'
    Rosetta = '213'
    glitchyspin = '302'
    System = '116'
    IOPowerSources = '141'
    PanicStats = '205'
    PowerLog_230 = '230'
    LongRunloopHang = '222'
    HomeProductsAnalytics = '311'
    DifferentialPrivacy_150 = '150'
    Rhodes = '214'
    ProactiveEventTrackerTransparency = '303'
    WiFi = '117'
    SymptomsCPUWakes = '142'
    SymptomsCPUUsageFatal = '206'
    Crash_109 = '109'
    ShortRunloopHang = '223'
    CoreHandwriting_231 = '231'
    ForceReset = '151'
    SiriAppSelection = '215'
    PrivateFederatedLearning = '304'
    Bluetooth = '118'
    SCPMotion = '143'
    HangSpin = '207'
    StepCount = '160'
    RTCTransparency = '224'
    DiagnosticRequest = '312'
    MemorySnapshot = '152'
    Rosetta_B = '216'
    AudioAccessory = '305'
    General = '119'
    HotSpotIOMicroSS = '144'
    GeoServicesTransparency = '233'
    MotionState = '161'
    AppStoreTransparency = '225'
    SiriSearchFeedback = '313'
    BearTrapReserved = '153'
    Portrait = '217'
    AWDMetricLog = 'metriclog'
    SymptomsIO = '145'
    SubmissionReserved = '170'
    WifiCrash = '209'
    Natalies = '162'
    SecurityTransparency = '226'
    BiomeMapReduce = '234'
    MemoryGraph = '154'
    MultichannelAudio = '218'
    honeybee_payload = '146'
    MesaReserved = '171'
    WifiSensing = '235'
    SiriMiss = '163'
    ExcResourceThreads_227 = '227'
    TestA = 'T01'
    NetworkUsage = '155'
    WifiReserved = '180'
    SiriActionPrediction = '219'
    honeybee_heartbeat = '147'
    ECCEvent = '172'
    KeyTransparency = '236'
    SubDiagHeartBeat = '164'
    ThirdPartyHang = '228'
    OSFault = '308'
    CoreTime = '156'
    WifiDriverReserved = '181'
    Crash_309 = '309'
    honeybee_issue = '148'
    CellularPerfReserved = '173'
    TestB = 'T02'
    StorageStatus = '165'
    SiriNotificationTransparency = '229'
    TestC = 'T03'
    CPUMicroSS = '157'
    AccessoryUpdate = '182'
    xprotect = '20'
    MultitouchFirmware = '149'
    MicroStackshot = '174'
    AppLaunchDiagnostics = '238'
    KeyboardAccuracy = '166'
    GPURestart = '21'
    FaceTime = '191'
    DuetKnowledgeCollector = '158'
    OTASUpdate = '183'
    ExcResourceThreads_327 = '327'
    ExcResource_22 = '22'
    DuetDB = '175'
    ThirdPartyHangDeveloper = '328'
    PrivacySettings = '167'
    GasGauge = '192'
    MicroStackShots = '23'
    BasebandCrash = '159'
    GPURestart_184 = '184'
    SystemWatchdogCrash = '409'
    FlashStatus = '176'
    SleepWakeFailure = '24'
    CarouselEvent = '168'
    AggregateD = '193'
    WakeupsMonitorViolation = '25'
    DifferentialPrivacy_50 = '50'
    ExcResource_185 = '185'
    UIAutomation = '177'
    ping = '26'
    SiriTransaction = '169'
    SURestore = '194'
    KtraceStackshot = '186'
    WirelessDiagnostics = '27'
    PowerLogLite = '178'
    SKAdNetworkAnalytics = '237'


class CrashReportBase:
    def __init__(self, metadata: Mapping, data: str, filename: str = None):
        self.filename = filename
        self._metadata = metadata
        self._data = data
        self._parse()

    def _parse(self):
        self._is_json = False
        try:
            self._data = json.loads(self._data)
            self._is_json = True
        except json.decoder.JSONDecodeError:
            pass

    @cached_property
    def bug_type(self) -> BugType:
        return BugType(self.bug_type_str)

    @cached_property
    def bug_type_str(self) -> str:
        return self._metadata['bug_type']

    @cached_property
    def incident_id(self):
        return self._metadata.get('incident_id')

    @cached_property
    def timestamp(self) -> datetime:
        timestamp = self._metadata.get('timestamp')
        timestamp_without_timezone = timestamp.rsplit(' ', 1)[0]
        return datetime.strptime(timestamp_without_timezone, '%Y-%m-%d %H:%M:%S.%f')

    @cached_property
    def name(self) -> str:
        return self._metadata.get('name')

    def __repr__(self) -> str:
        filename = ''
        if self.filename:
            filename = f'FILENAME:{posixpath.basename(self.filename)} '
        return f'<{self.__class__} {filename}TIMESTAMP:{self.timestamp}>'

    def __str__(self) -> str:
        filename = ''
        if self.filename:
            filename = self.filename

        return click.style(f'{self.incident_id} {self.timestamp}\n{filename}\n\n', fg='cyan')


class UserModeCrashReport(CrashReportBase):
    def _parse_field(self, name: str) -> str:
        name += ':'
        for line in self._data.split('\n'):
            if line.startswith(name):
                field = line.split(name, 1)[1]
                field = field.strip()
                return field

    @cached_property
    def faulting_thread(self) -> int:
        if self._is_json:
            return self._data['faultingThread']
        else:
            return int(self._parse_field('Triggered by Thread'))

    @cached_property
    def frames(self) -> List[Frame]:
        result = []
        if self._is_json:
            thread_index = self.faulting_thread
            images = self._data['usedImages']
            for frame in self._data['threads'][thread_index]['frames']:
                image = images[frame['imageIndex']]
                result.append(
                    Frame(image_name=image.get('path'), image_base=image.get('base'), symbol=frame.get('symbol'),
                          image_offset=frame.get('imageOffset'), symbol_offset=frame.get('symbolLocation')))
        else:
            in_frames = False
            for line in self._data.split('\n'):
                if in_frames:
                    splitted = line.split()

                    if len(splitted) == 0:
                        break

                    assert splitted[-2] == '+'
                    image_base = splitted[-3]
                    if image_base.startswith('0x'):
                        result.append(Frame(image_name=splitted[1], image_base=int(image_base, 16), symbol=None,
                                            image_offset=int(splitted[-1]), symbol_offset=None))
                    else:
                        # symbolicated
                        result.append(Frame(image_name=splitted[1], image_base=None, symbol=image_base,
                                            image_offset=None, symbol_offset=int(splitted[-1])))

                if line.startswith(f'Thread {self.faulting_thread} Crashed:'):
                    in_frames = True

        return result

    @cached_property
    def registers(self) -> List[Register]:
        result = []
        if self._is_json:
            thread_index = self._data['faultingThread']
            thread_state = self._data['threads'][thread_index]['threadState']

            if 'x' in thread_state:
                for i, reg_x in enumerate(thread_state['x']):
                    result.append(Register(name=f'x{i}', value=reg_x['value']))

            for i, (name, value) in enumerate(thread_state.items()):
                if name == 'x':
                    for j, reg_x in enumerate(value):
                        result.append(Register(name=f'x{j}', value=reg_x['value']))
                else:
                    if isinstance(value, dict):
                        result.append(Register(name=name, value=value['value']))
        else:
            in_frames = False
            for line in self._data.split('\n'):
                if in_frames:
                    splitted = line.split()

                    if len(splitted) == 0:
                        break

                    for i in range(0, len(splitted), 2):
                        register_name = splitted[i]
                        if not register_name.endswith(':'):
                            break

                        register_name = register_name[:-1]
                        register_value = int(splitted[i + 1], 16)

                        result.append(Register(name=register_name, value=register_value))

                if line.startswith(f'Thread {self.faulting_thread} crashed with ARM Thread State'):
                    in_frames = True

        return result

    @cached_property
    def exception_type(self):
        if self._is_json:
            return self._data['exception'].get('type')
        else:
            return self._parse_field('Exception Type')

    @cached_property
    def exception_subtype(self) -> Optional[str]:
        if self._is_json:
            return self._data['exception'].get('subtype')
        else:
            return self._parse_field('Exception Subtype')

    @cached_property
    def application_specific_information(self) -> Optional[str]:
        result = ''
        if self._is_json:
            asi = self._data.get('asi')
            if asi is None:
                return None
            return asi
        else:
            in_frames = False
            for line in self._data.split('\n'):
                if in_frames:
                    line = line.strip()
                    if len(line) == 0:
                        break

                    result += line + '\n'

                if line.startswith('Application Specific Information:'):
                    in_frames = True

        result = result.strip()
        if not result:
            return None
        return result

    def __str__(self) -> str:
        result = super().__str__()
        result += click.style(f'Exception: {self.exception_type}\n', bold=True)

        if self.exception_subtype:
            result += click.style('Exception Subtype: ', bold=True)
            result += f'{self.exception_subtype}\n'

        if self.application_specific_information:
            result += click.style('Application Specific Information: ', bold=True)
            result += str(self.application_specific_information)

        result += '\n'

        result += click.style('Registers:', bold=True)
        for i, register in enumerate(self.registers):
            if i % 4 == 0:
                result += '\n'

            result += f'{register.name} = 0x{register.value:016x} '.rjust(30)

        result += '\n\n'

        result += click.style('Frames:\n', bold=True)
        for frame in self.frames:
            image_base = '_HEADER'
            if frame.image_base is not None:
                image_base = f'0x{frame.image_base:x}'
            result += f'\t[{frame.image_name}] {image_base}'
            if frame.image_offset:
                result += f' + 0x{frame.image_offset:x}'
            if frame.symbol is not None:
                result += f' ({frame.symbol} + 0x{frame.symbol_offset:x})'
            result += '\n'

        return result


class KernelModeCrashReport(KernelPanic, CrashReportBase):
    def __init__(self, metadata: Mapping, data: str, filename: str = None):
        CrashReportBase.__init__(self, metadata, data, filename=filename)
        KernelPanic.__init__(self, metadata, data, filename=filename)

    @cached_property
    def bug_type(self) -> BugType:
        return BugType(self._metadata['bug_type'])


def get_crash_report_from_file(crash_report_file: IO) -> CrashReportBase:
    metadata = json.loads(crash_report_file.readline())

    try:
        bug_type = BugType(metadata['bug_type'])
    except ValueError:
        return CrashReportBase(metadata, crash_report_file.read(), crash_report_file.name)

    bug_type_parsers = {
        BugType.ForceReset: KernelModeCrashReport,
        BugType.Panic_210: KernelModeCrashReport,
        BugType.Crash_109: UserModeCrashReport,
        BugType.Crash_309: UserModeCrashReport,
        BugType.ExcResourceThreads_327: UserModeCrashReport,
        BugType.ExcResource_385: UserModeCrashReport,
    }

    parser = bug_type_parsers.get(bug_type)
    if parser is None:
        return CrashReportBase(metadata, crash_report_file.read(), crash_report_file.name)

    return parser(metadata, crash_report_file.read(), crash_report_file.name)


def get_crash_report_from_buf(crash_report_buf: str, filename: str = None) -> CrashReportBase:
    file = StringIO(crash_report_buf)
    file.name = filename
    return get_crash_report_from_file(file)
